# pylint: disable=missing-class-docstring missing-module-docstring
import unittest
from dataclasses import dataclass

from liquid.expression import BooleanExpression
from liquid.expression import ConditionalExpression
from liquid.expression import Expression
from liquid.expression import FilteredExpression
from liquid.expression import InfixExpression
from liquid.expression import PrefixExpression
from liquid.expression import IdentifierPathElement
from liquid.expression import Identifier
from liquid.expression import Boolean
from liquid.expression import Filter
from liquid.expression import StringLiteral
from liquid.expression import IntegerLiteral
from liquid.expression import FloatLiteral
from liquid.expression import RangeLiteral
from liquid.expression import TRUE

from liquid.expressions import parse_filtered_expression
from liquid.expressions import parse_conditional_expression
from liquid.expressions import parse_conditional_expression_with_parens

from liquid.exceptions import LiquidSyntaxError


@dataclass
class Case:
    description: str
    expression: str
    expect: Expression


class ParseFilteredExpressionTestCase(unittest.TestCase):
    test_cases = [
        Case(
            "string literal single quotes",
            "'foobar'",
            FilteredExpression(
                expression=StringLiteral("foobar"),
                filters=[],
            ),
        ),
        Case(
            "string literal double quotes",
            '"foobar"',
            FilteredExpression(
                expression=StringLiteral("foobar"),
                filters=[],
            ),
        ),
        Case(
            "integer literal",
            "7",
            FilteredExpression(
                expression=IntegerLiteral(7),
                filters=[],
            ),
        ),
        Case(
            "negative integer literal statement expression",
            "-7",
            FilteredExpression(
                expression=IntegerLiteral(-7),
                filters=[],
            ),
        ),
        Case(
            "float literal statement expression",
            "3.14",
            FilteredExpression(
                expression=FloatLiteral(3.14),
                filters=[],
            ),
        ),
        Case(
            "negative float literal statement expression",
            "-3.14",
            FilteredExpression(
                expression=FloatLiteral(-3.14),
                filters=[],
            ),
        ),
        Case(
            "literal true",
            "true",
            FilteredExpression(
                expression=Boolean(True),
                filters=[],
            ),
        ),
        Case(
            "literal false",
            "false",
            FilteredExpression(
                expression=Boolean(False),
                filters=[],
            ),
        ),
        Case(
            "range expression with integer start and stop",
            "(1..4)",
            FilteredExpression(
                expression=RangeLiteral(IntegerLiteral(1), IntegerLiteral(4)),
                filters=[],
            ),
        ),
        Case(
            "range expression with identifier start and stop",
            "(a..b.c)",
            FilteredExpression(
                expression=RangeLiteral(
                    Identifier(
                        path=[
                            IdentifierPathElement("a"),
                        ]
                    ),
                    Identifier(
                        path=[
                            IdentifierPathElement("b"),
                            IdentifierPathElement("c"),
                        ]
                    ),
                ),
                filters=[],
            ),
        ),
        Case(
            "single identifier statement expression",
            "collection",
            FilteredExpression(
                expression=Identifier(
                    path=[IdentifierPathElement("collection")],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier",
            "collection.products",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier by double quoted key",
            'collection["products"]',
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier by double quoted key followed by index",
            'collection["products"][0]',
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                        IdentifierPathElement(0),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier by double quoted key followed by identifier",
            'collection["products"].title',
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                        IdentifierPathElement("title"),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier by single quoted key",
            "collection['products']",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier with array index",
            "collection.products[0]",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                        IdentifierPathElement(0),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier with array index from variable",
            "collection.products[i]",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                        Identifier(
                            path=[IdentifierPathElement("i")],
                        ),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier with array index from chained identifier",
            "collection.products[some.['object'].count]",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("products"),
                        Identifier(
                            path=[
                                IdentifierPathElement("some"),
                                IdentifierPathElement("object"),
                                IdentifierPathElement("count"),
                            ],
                        ),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "chained identifier with object from chained identifier",
            "linklists[section.settings.menu]",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("linklists"),
                        Identifier(
                            path=[
                                IdentifierPathElement("section"),
                                IdentifierPathElement("settings"),
                                IdentifierPathElement("menu"),
                            ],
                        ),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            (
                "chained identifier with object from chained identifier "
                "and trailing identifier"
            ),
            "linklists[section.settings.menu].links",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("linklists"),
                        Identifier(
                            path=[
                                IdentifierPathElement("section"),
                                IdentifierPathElement("settings"),
                                IdentifierPathElement("menu"),
                            ],
                        ),
                        IdentifierPathElement("links"),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "string literal with filter",
            "'foo' | upcase",
            FilteredExpression(
                expression=StringLiteral("foo"),
                filters=[Filter(name="upcase", args=[])],
            ),
        ),
        Case(
            "string literal with filter and filter arguments",
            "'foo' | append: 'bar'",
            FilteredExpression(
                expression=StringLiteral("foo"),
                filters=[Filter(name="append", args=[StringLiteral("bar")])],
            ),
        ),
        Case(
            "identifier with filter",
            "collection.title | upcase",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("collection"),
                        IdentifierPathElement("title"),
                    ],
                ),
                filters=[Filter(name="upcase", args=[])],
            ),
        ),
        Case(
            "integer literal with filter and integer argument",
            "4 | at_least: 5",
            FilteredExpression(
                expression=IntegerLiteral(4),
                filters=[Filter(name="at_least", args=[IntegerLiteral(5)])],
            ),
        ),
        Case(
            "float literal with filter and float argument",
            "4.1 | divided_by: 5.2",
            FilteredExpression(
                expression=FloatLiteral(4.1),
                filters=[Filter(name="divided_by", args=[FloatLiteral(5.2)])],
            ),
        ),
        Case(
            "string literal with filter and string argument",
            "'foo' | append: 'bar'",
            FilteredExpression(
                expression=StringLiteral("foo"),
                filters=[Filter(name="append", args=[StringLiteral("bar")])],
            ),
        ),
        Case(
            "string literal with filter and identifier argument",
            "'foo' | append: collection.title",
            FilteredExpression(
                expression=StringLiteral("foo"),
                filters=[
                    Filter(
                        name="append",
                        args=[
                            Identifier(
                                path=[
                                    IdentifierPathElement("collection"),
                                    IdentifierPathElement("title"),
                                ],
                            )
                        ],
                    )
                ],
            ),
        ),
        Case(
            "string literal with filter and multiple arguments",
            '"Liquid" | slice: 2, 5',
            FilteredExpression(
                expression=StringLiteral("Liquid"),
                filters=[
                    Filter(
                        name="slice",
                        args=[
                            IntegerLiteral(2),
                            IntegerLiteral(5),
                        ],
                    )
                ],
            ),
        ),
        Case(
            "string literal with multiple filters",
            '"Liquid" | slice: 2, 5 | upcase',
            FilteredExpression(
                expression=StringLiteral("Liquid"),
                filters=[
                    Filter(
                        name="slice",
                        args=[
                            IntegerLiteral(2),
                            IntegerLiteral(5),
                        ],
                    ),
                    Filter(name="upcase", args=[]),
                ],
            ),
        ),
        Case(
            "inconsistent whitespace",
            ' "Liquid"   |slice: 2,5',
            FilteredExpression(
                expression=StringLiteral("Liquid"),
                filters=[
                    Filter(
                        name="slice",
                        args=[
                            IntegerLiteral(2),
                            IntegerLiteral(5),
                        ],
                    )
                ],
            ),
        ),
        Case(
            "negative array index",
            "products[-1]",
            FilteredExpression(
                expression=Identifier(
                    path=[
                        IdentifierPathElement("products"),
                        IdentifierPathElement(-1),
                    ],
                ),
                filters=[],
            ),
        ),
        Case(
            "filter with a named argument",
            "'Liquid' | default: 'hello', allow_false: true",
            FilteredExpression(
                expression=StringLiteral("Liquid"),
                filters=[
                    Filter(
                        name="default",
                        args=[StringLiteral("hello")],
                        kwargs={"allow_false": Boolean(True)},
                    )
                ],
            ),
        ),
    ]

    def test_parse_filtered_expression(self):
        """Test that we can parse liquid statement expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_filtered_expression(case.expression)
                self.assertEqual(expr, case.expect)

    def test_parse_conditional_expression(self):
        """Test that the non-standard conditional expression parser is backwards
        compatible with standard filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_conditional_expression(case.expression)
                self.assertEqual(expr, case.expect)

    def test_parse_conditional_expression_with_parens(self):
        """Test that the non-standard conditional expression parser is backwards
        compatible with standard filtered expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_conditional_expression_with_parens(case.expression)
                self.assertEqual(expr, case.expect)

    def test_double_pipe(self):
        """Test that a two consecutive pipe characters raises a syntax error."""
        expr = r'"failure" || upcase'
        with self.assertRaises(LiquidSyntaxError) as raised:
            parse_filtered_expression(expr)
        self.assertEqual(
            str(raised.exception),
            "unexpected pipe or missing filter name, on line 1",
        )

    def test_double_comma(self):
        """Test that two consecutive comma characters raises a syntax error."""
        expr = r'"hello" | slice: 2,, 4'
        with self.assertRaises(LiquidSyntaxError) as raised:
            parse_filtered_expression(expr)
        self.assertEqual(str(raised.exception), "unexpected ',', on line 1")

    def test_double_colon(self):
        """Test that two consecutive colon characters raises a syntax error."""
        expr = r'"hello" | slice:: 2'
        with self.assertRaises(LiquidSyntaxError) as raised:
            parse_filtered_expression(expr)
        self.assertEqual(str(raised.exception), "unexpected ':', on line 1")

    def test_invalid_filter_name(self):
        """Test that an invalid filter name raises a syntax error."""
        expr = r'"hello" | £%1 '
        with self.assertRaises(LiquidSyntaxError) as raised:
            parse_filtered_expression(expr)
        self.assertEqual(str(raised.exception), "unexpected '£', on line 1")

    def test_missing_comma(self):
        """Test that a missing comma raises a syntax error."""
        expr = r'"hello" | slice: 1 3 '
        with self.assertRaises(LiquidSyntaxError) as raised:
            parse_filtered_expression(expr)
        self.assertEqual(
            str(raised.exception),
            "expected ',', found 'integer', on line 1",
        )

    def test_missing_colon(self):
        """Test that a missing colon raises a syntax error."""
        expr = r'"hello" | slice 1, 3 '
        with self.assertRaises(LiquidSyntaxError) as raised:
            parse_filtered_expression(expr)
        self.assertEqual(
            str(raised.exception), "expected ':', found 'integer', on line 1"
        )


class ParseConditionalExpressionTestCase(unittest.TestCase):
    """Test cases for the non-standard filtered expressions with inline conditions."""

    test_cases = [
        Case(
            description="string literal with condition",
            expression="'foo' if true",
            expect=ConditionalExpression(
                expression=FilteredExpression(StringLiteral("foo")),
                condition=BooleanExpression(expression=TRUE),
            ),
        ),
        Case(
            description="string literal with condition and alternative",
            expression="'foo' if true else 'bar'",
            expect=ConditionalExpression(
                expression=FilteredExpression(StringLiteral("foo")),
                condition=BooleanExpression(expression=TRUE),
                alternative=FilteredExpression(StringLiteral("bar")),
            ),
        ),
        Case(
            description="identifiers with condition and alternative",
            expression="products[0] if products else 'bar'",
            expect=ConditionalExpression(
                expression=FilteredExpression(
                    Identifier(
                        path=[
                            IdentifierPathElement("products"),
                            IdentifierPathElement(0),
                        ]
                    )
                ),
                condition=BooleanExpression(
                    expression=Identifier(path=[IdentifierPathElement("products")])
                ),
                alternative=FilteredExpression(StringLiteral("bar")),
            ),
        ),
        Case(
            description="string literal with condition, alternative and filter",
            expression="'foo' | upcase if true else 'bar'",
            expect=ConditionalExpression(
                expression=FilteredExpression(
                    StringLiteral("foo"),
                    filters=[Filter(name="upcase", args=[])],
                ),
                condition=BooleanExpression(expression=TRUE),
                alternative=FilteredExpression(StringLiteral("bar")),
                filters=[],
            ),
        ),
        Case(
            description=("string literal with condition and alternative filter"),
            expression="'foo' if true else 'bar' | upcase",
            expect=ConditionalExpression(
                expression=FilteredExpression(StringLiteral("foo")),
                condition=BooleanExpression(expression=TRUE),
                alternative=FilteredExpression(
                    StringLiteral("bar"),
                    filters=[Filter(name="upcase", args=[])],
                ),
                filters=[],
            ),
        ),
        Case(
            description=("string literal with condition, alternative and tail filter"),
            expression="'foo' if true else 'bar' || append: 'baz' | upcase",
            expect=ConditionalExpression(
                expression=FilteredExpression(StringLiteral("foo")),
                condition=BooleanExpression(expression=TRUE),
                alternative=FilteredExpression(StringLiteral("bar")),
                filters=[
                    Filter(name="append", args=[StringLiteral("baz")]),
                    Filter(name="upcase", args=[]),
                ],
            ),
        ),
        Case(
            description=("tail filter with keyword argument"),
            expression="'foo' if true else 'bar' || default: 'baz', allow_false:true",
            expect=ConditionalExpression(
                expression=FilteredExpression(StringLiteral("foo")),
                condition=BooleanExpression(expression=TRUE),
                alternative=FilteredExpression(StringLiteral("bar")),
                filters=[
                    Filter(
                        name="default",
                        args=[StringLiteral("baz")],
                        kwargs={"allow_false": TRUE},
                    )
                ],
            ),
        ),
    ]

    def test_parse_conditional_expression(self):
        """Test that we can parse conditional expressions."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_conditional_expression(case.expression)
                self.assertEqual(expr, case.expect)

    def test_parse_conditional_expression_with_parens(self):
        """Test that the other conditional expression parser is backwards
        compatible with conditional expressions that don't support parentheses."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_conditional_expression_with_parens(case.expression)
                self.assertEqual(expr, case.expect)


class ParseConditionalNotExpressionTestCase(unittest.TestCase):
    """Test cases for the non-standard filtered expressions with inline conditions that
    support logical `not` and grouping with parentheses."""

    test_cases = [
        Case(
            description="string literal with condition",
            expression="'foo' if not true",
            expect=ConditionalExpression(
                expression=FilteredExpression(StringLiteral("foo")),
                condition=BooleanExpression(
                    expression=PrefixExpression(
                        "not",
                        right=Boolean(True),
                    ),
                ),
            ),
        ),
        Case(
            description="string literal with condition and alternative",
            expression="'foo' if (true and false and false) or true else 'bar'",
            expect=ConditionalExpression(
                expression=FilteredExpression(StringLiteral("foo")),
                condition=BooleanExpression(
                    expression=InfixExpression(
                        left=InfixExpression(
                            left=Boolean(True),
                            operator="and",
                            right=InfixExpression(
                                left=Boolean(False),
                                operator="and",
                                right=Boolean(False),
                            ),
                        ),
                        operator="or",
                        right=Boolean(True),
                    ),
                ),
                alternative=FilteredExpression(StringLiteral("bar")),
            ),
        ),
    ]

    def test_parse_conditional_expression_with_parens(self):
        """Test that the other conditional expression parser is backwards
        compatible with conditional expressions that don't support parentheses."""
        for case in self.test_cases:
            with self.subTest(msg=case.description):
                expr = parse_conditional_expression_with_parens(case.expression)
                self.assertEqual(expr, case.expect)



"""Test Python Liquid's async API."""
# pylint: disable=missing-class-docstring
import asyncio
import tempfile
import time
import unittest

from collections import abc
from pathlib import Path

from typing import NamedTuple
from typing import Dict

# assert_awaited* were new in Python 3.8, so we're using the backport.
from mock import patch

from liquid import Template
from liquid import Environment
from liquid import FileSystemLoader

from liquid.loaders import TemplateSource
from liquid.loaders import ChoiceLoader
from liquid.loaders import DictLoader

from liquid.exceptions import TemplateNotFound
from liquid.template import BoundTemplate


class Case(NamedTuple):
    """Table driven test helper."""

    description: str
    template: str
    context: dict
    expect: str
    template_name: str
    calls: int


class LoadAsyncTestCase(unittest.TestCase):
    """Test that we can load templates asynchronously."""

    def test_basic_async(self):
        """Test that we can render a template asynchronously."""
        template = Template("{% for x in (1..3) %}{{x}}-{% endfor %}")

        async def coro():
            return await template.render_async()

        result = asyncio.run(coro())
        self.assertEqual(result, "1-2-3-")

    def test_load_template_async(self):
        """Test that we can load a template asynchronously."""
        env = Environment(loader=FileSystemLoader("tests/fixtures/dropify/"))

        async def coro():
            template = await env.get_template_async("index.liquid")
            self.assertIsInstance(template, BoundTemplate)

        asyncio.run(coro())

    def test_template_not_found_async(self):
        """Test that non-existent templates raise TemplateNotFound."""
        env = Environment(loader=FileSystemLoader("tests/fixtures/dropify/"))

        async def coro():
            return await env.get_template_async("nosuchthing.liquid")

        with self.assertRaises(TemplateNotFound):
            asyncio.run(coro())

    def test_cached_template_async(self):
        """Test that async loaded templates are cached."""
        env = Environment(loader=FileSystemLoader("tests/fixtures/dropify/"))

        async def coro():
            return await env.get_template_async("index.liquid")

        with patch(
            "liquid.loaders.FileSystemLoader.get_source_async", autospec=True
        ) as source:
            source.side_effect = [
                TemplateSource(
                    r"hello {{ you }}",
                    "some_template",
                    None,
                )
            ]
            template = asyncio.run(coro())
            self.assertIsInstance(template, BoundTemplate)
            source.assert_awaited_once()

            # Get the same template again.
            template = asyncio.run(coro())
            self.assertIsInstance(template, BoundTemplate)
            self.assertEqual(source.call_count, 1)

    def test_async_up_to_date(self):
        """Test that async uptodate functions are awaited."""
        with tempfile.TemporaryDirectory() as tmpdir:
            template_path = Path(tmpdir) / "some.liquid"

            # Initial template content
            with template_path.open("w", encoding="UTF-8") as fd:
                fd.write("hello there\n")

            env = Environment(
                loader=FileSystemLoader(search_path=tmpdir),
                auto_reload=True,
            )

            async def coro():
                template = await env.get_template_async("some.liquid")
                self.assertIsInstance(template, BoundTemplate)
                uptodate = await template.is_up_to_date_async()
                self.assertTrue(uptodate)

                # Update template content.
                time.sleep(0.01)  # Make sure some time has passed.
                template_path.touch()

                # Template is not up to date.
                uptodate = await template.is_up_to_date_async()
                self.assertFalse(uptodate)

                template = await env.get_template_async("some.liquid")
                self.assertIsInstance(template, BoundTemplate)

                # Template is up to date.
                uptodate = await template.is_up_to_date_async()
                self.assertTrue(uptodate)

            asyncio.run(coro())

    def test_nested_include_async(self):
        """Test that nested includes are rendered asynchronously."""

        async def coro(template: BoundTemplate):
            return await template.render_async()

        with patch(
            "liquid.loaders.DictLoader.get_source_async", autospec=True
        ) as source:
            source.side_effect = [
                TemplateSource(
                    "{% include 'bar' %}",
                    "bar",
                    None,
                ),
                TemplateSource(
                    "{% for x in (1..3) %}{{x}}-{% endfor %}",
                    "foo",
                    None,
                ),
            ]

            env = Environment()
            template = env.from_string("{% include 'foo' %}")

            result = asyncio.run(coro(template))

            source.assert_awaited()
            self.assertEqual(source.await_count, 2)
            self.assertEqual(result, "1-2-3-")

    def test_include_async(self):
        """Test that included templates are rendered asynchronously"""
        test_cases = [
            Case(
                description="simple include",
                template="{% include 'foo' %}",
                context={},
                expect="1-2-3-",
                template_name="foo",
                calls=1,
            ),
            Case(
                description="include with arguments",
                template="{% include 'foo' with foo as bar, baz: 'hello' %}",
                context={},
                expect="1-2-3-",
                template_name="foo",
                calls=1,
            ),
            Case(
                description="include for array",
                template="{% include 'foo' for array %}",
                context={"array": [1, 2]},
                expect="1-2-3-1-2-3-",
                template_name="foo",
                calls=1,  # cached
            ),
        ]

        async def coro(template: BoundTemplate):
            return await template.render_async()

        for case in test_cases:
            with self.subTest(msg=case.description):
                env = Environment()
                template = env.from_string(case.template, globals=case.context)

                with patch(
                    "liquid.loaders.DictLoader.get_source_async", autospec=True
                ) as source:
                    source.side_effect = [
                        TemplateSource(
                            "{% for x in (1..3) %}{{x}}-{% endfor %}",
                            "foo",
                            None,
                        )
                    ]

                    result = asyncio.run(coro(template))

                    source.assert_awaited_with(env.loader, env=env, template_name="foo")
                    self.assertEqual(source.call_count, case.calls)
                    self.assertEqual(result, case.expect)

    def test_cover_render_async(self):
        """Test that `render`ed templates are loaded asynchronously."""
        test_cases = [
            Case(
                description="simple render",
                template="{% render 'foo' %}",
                context={},
                expect="1-2-3-",
                template_name="foo",
                calls=1,
            ),
            Case(
                description="render with arguments",
                template="{% render 'foo' with foo as bar, baz: 'hello' %}",
                context={},
                expect="1-2-3-",
                template_name="foo",
                calls=1,
            ),
            Case(
                description="render for array",
                template="{% render 'foo' for array %}",
                context={"array": [1, 2]},
                expect="1-2-3-1-2-3-",
                template_name="foo",
                calls=1,  # cached
            ),
        ]

        async def coro(template: BoundTemplate):
            return await template.render_async()

        for case in test_cases:
            with self.subTest(msg=case.description):
                with patch(
                    "liquid.loaders.DictLoader.get_source_async", autospec=True
                ) as source:
                    source.side_effect = [
                        TemplateSource(
                            "{% for x in (1..3) %}{{x}}-{% endfor %}",
                            "foo",
                            None,
                        )
                    ]

                    env = Environment()
                    template = env.from_string(case.template, globals=case.context)

                    result = asyncio.run(coro(template))

                    source.assert_awaited_with(env.loader, env=env, template_name="foo")
                    self.assertEqual(source.call_count, case.calls)
                    self.assertEqual(result, case.expect)


class MockAsyncDrop(abc.Mapping):
    def __init__(self, val):
        self.key = "foo"
        self.val = val

        self.call_count = 0
        self.await_count = 0

    def __len__(self):
        return 1

    def __iter__(self):
        return iter([self.key])

    def __getitem__(self, k):
        self.call_count += 1
        if k == self.key:
            return self.val
        raise KeyError(k)

    async def __getitem_async__(self, k):
        # Do IO here
        self.await_count += 1
        if k == self.key:
            return self.val
        raise KeyError(k)


class AsyncDropTestCase(unittest.TestCase):
    """Test that we can get items from a drop asynchronously."""

    def setUp(self) -> None:
        self.drop = MockAsyncDrop("hello")

    def test_awaited(self):
        """Test that __getitem_async__ is awaited when using render_async."""
        env = Environment()
        template = env.from_string(r"{{ drop.foo }}")

        async def coro():
            return await template.render_async(drop=self.drop)

        result = asyncio.run(coro())

        self.assertEqual(result, "hello")
        self.assertEqual(self.drop.await_count, 1)
        self.assertEqual(self.drop.call_count, 0)

    def test_called(self):
        """Test that __getitem_async__ is not called when rendering synchronously."""
        env = Environment()
        template = env.from_string(r"{{ drop.foo }}")
        result = template.render(drop=self.drop)

        self.assertEqual(result, "hello")
        self.assertEqual(self.drop.await_count, 0)
        self.assertEqual(self.drop.call_count, 1)


class AsyncMatterDictLoader(DictLoader):
    def __init__(
        self,
        templates: Dict[str, str],
        matter: Dict[str, Dict[str, object]],
    ):
        super().__init__(templates)
        self.matter = matter

    async def get_source_async(
        self, _: Environment, template_name: str
    ) -> TemplateSource:
        try:
            source = self.templates[template_name]
        except KeyError as err:
            raise TemplateNotFound(template_name) from err

        return TemplateSource(
            source=source,
            filename=template_name,
            uptodate=None,
            matter=self.matter.get(template_name),
        )


class MatterLoaderTestCase(unittest.TestCase):
    def test_matter_loader(self):
        """Test that template loaders can add to render context."""
        loader = AsyncMatterDictLoader(
            templates={
                "some": "Hello, {{ you }}{{ username }}!",
                "other": "Goodbye, {{ you }}{{ username }}.",
                "thing": "{{ you }}{{ username }}",
            },
            matter={
                "some": {"you": "World"},
                "other": {"username": "Smith"},
            },
        )

        env = Environment(loader=loader)

        async def coro():
            template = await env.get_template_async("some")
            self.assertEqual(await template.render_async(), "Hello, World!")

            template = await env.get_template_async("other")
            self.assertEqual(await template.render_async(), "Goodbye, Smith.")

            template = await env.get_template_async("thing")
            self.assertEqual(await template.render_async(), "")

        asyncio.run(coro())

    def test_matter_global_priority(self):
        """Test that matter variables take priority over globals."""
        loader = AsyncMatterDictLoader(
            templates={"some": "Hello, {{ you }}!"},
            matter={"some": {"you": "Liquid"}},
        )

        env = Environment(loader=loader, globals={"you": "World"})

        async def coro():
            template = await env.get_template_async("some", globals={"you": "Jinja"})
            self.assertEqual(await template.render_async(), "Hello, Liquid!")

        asyncio.run(coro())

    def test_matter_local_priority(self):
        """Test that render args take priority over matter variables."""
        loader = AsyncMatterDictLoader(
            templates={"some": "Hello, {{ you }}!"},
            matter={"some": {"you": "Liquid"}},
        )

        env = Environment(loader=loader)

        async def coro():
            template = await env.get_template_async("some")
            self.assertEqual(await template.render_async(you="John"), "Hello, John!")

        asyncio.run(coro())


class ChoiceLoaderTestCase(unittest.TestCase):
    def test_choose_between_loaders(self):
        """Test that we can load templates from a list of loaders."""
        loader = ChoiceLoader(
            loaders=[
                DictLoader({"a": "Hello, {{ you }}!"}),
                DictLoader(
                    {
                        "a": "unreachable",
                        "b": "the quick brown {{ animal | default: 'fox' }}",
                    }
                ),
            ]
        )

        env = Environment(loader=loader)

        async def coro():
            template = await env.get_template_async("a")
            self.assertEqual(await template.render_async(you="World"), "Hello, World!")

            template = await env.get_template_async("b")
            self.assertEqual(await template.render_async(), "the quick brown fox")

            with self.assertRaises(TemplateNotFound):
                await env.get_template_async("c")

        asyncio.run(coro())



"""Golden tests cases for testing liquid's `echo` tag."""

import datetime
from liquid.golden.case import Case

cases = [
    Case(
        description="render a string literal",
        template=r"{% echo 'hello' %}",
        expect="hello",
    ),
    Case(
        description="render an integer literal",
        template=r"{% echo 123 %}",
        expect="123",
    ),
    Case(
        description="render a float literal",
        template=r"{% echo 1.23 %}",
        expect="1.23",
    ),
    Case(
        description="render a variable from the global namespace",
        template=r"{% echo product.title %}",
        expect="foo",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="render a variable from the local namespace",
        template=r"{% assign name = 'Brian' %}{% echo name %}",
        expect="Brian",
    ),
    Case(
        description="render an undefined variable",
        template=r"{% echo age %}",
        expect="",
    ),
    Case(
        description="render an undefined property",
        template=r"{% echo product.age %}",
        expect="",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="access an array item by index",
        template=r"{% echo product.tags[1] %}",
        expect="garden",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="access an array item by negative index",
        template=r"{% echo product.tags[-2] %}",
        expect="sports",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="access array item by index stored in a local variable",
        template=r"{% assign i = 1 %}{% echo product.tags[i] %}",
        expect="garden",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="render a global identifier with a filter",
        template=r"{% echo product.title | upcase %}",
        expect="FOO",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="dump an array from the global context",
        template=r"{% echo product.tags %}",
        expect="sportsgarden",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="assign a variable the value of an existing variable",
        template=(
            r"{% capture some %}hello{% endcapture %}"
            r"{% assign other = some %}"
            r"{% assign some = 'foo' %}"
            r"{% echo some %}-{% echo other %}"
        ),
        expect="foo-hello",
    ),
    Case(
        description="traverse variables with bracketed identifiers",
        template=r"{% echo site.data.menu[include.menu][include.locale] %}",
        expect="it works!",
        globals={
            "site": {"data": {"menu": {"foo": {"bar": "it works!"}}}},
            "include": {"menu": "foo", "locale": "bar"},
        },
    ),
    Case(
        description="render the special built in variable 'now'",
        template=r"{% echo now | date: '%d/%m/%Y' %}",
        expect=datetime.datetime.now().strftime(r"%d/%m/%Y"),
        standard=False,
    ),
    Case(
        description="render the special built in variable 'today'",
        template=r"{% echo today | date: '%d/%m/%Y' %}",
        expect=datetime.date.today().strftime(r"%d/%m/%Y"),
        standard=False,
    ),
    Case(
        description="access an undefined variable by index",
        template=r"{% echo nosuchthing[0] %}",
        expect="",
    ),
    Case(
        description="nothing to echo",
        template=r"{% echo %}",
        expect="",
    ),
]



"""Golden tests cases for testing liquid's built-in `first` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="range literal last filter left value",
        template=r"{{ (1..3) | last }}",
        expect="3",
    ),
    Case(
        description="array of strings",
        template=r"{{ arr | last }}",
        expect="b",
        globals={"arr": ["a", "b"]},
    ),
    Case(
        description="array of things",
        template=r"{{ arr | last }}",
        expect="{}",
        globals={"arr": ["a", "b", 1, [], {}]},
    ),
    Case(
        description="empty array",
        template=r"{{ arr | last }}",
        expect="",
        globals={"arr": []},
    ),
    Case(
        description="left value not an array",
        template=r"{{ arr | last }}",
        expect="",
        globals={"arr": 12},
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | last }}",
        expect="",
    ),
    Case(
        description="last of a string",
        template=r"{{ 'hello' | last }}",
        expect="",
    ),
    Case(
        description="last of a hash",
        template=r"{{ a | last }}",
        expect="",
        globals={"a": {"b": 1, "c": 2}},
    ),
]



"""Common parse tree nodes."""

from abc import ABC
from abc import abstractmethod

from typing import Collection
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import TextIO

from typing_extensions import Literal

from liquid.context import Context
from liquid.expression import Expression

from liquid.token import Token, TOKEN_TAG, TOKEN_ILLEGAL

from liquid.exceptions import DisabledTagError
from liquid.exceptions import Error

IllegalToken = Token(-1, TOKEN_ILLEGAL, "")


class Node(ABC):
    """Base class for all nodes in a parse tree."""

    __slots__ = ()

    # Indicates that nodes that do automatic whitespace suppression should output this
    # node regardless of its contents.
    force_output = False

    def token(self) -> Token:
        """The token that started this node."""
        token: Token = getattr(self, "tok", IllegalToken)
        return token

    def raise_for_disabled(self, disabled_tags: Collection[str]) -> None:
        """Raise a DisabledTagError if this node's type is in the given list."""
        tok = self.token()
        if tok.type == TOKEN_TAG and tok.value in disabled_tags:
            raise DisabledTagError(
                f"{tok.value} usage is not allowed in this context",
                linenum=tok.linenum,
            )

    def render(self, context: Context, buffer: TextIO) -> Optional[bool]:
        """Check disabled tags before delegating to `render_to_output`."""
        if context.disabled_tags:
            self.raise_for_disabled(context.disabled_tags)
        return self.render_to_output(context, buffer)

    async def render_async(self, context: Context, buffer: TextIO) -> Optional[bool]:
        """An async version of :meth:`liquid.ast.Node.render`."""
        if context.disabled_tags:
            self.raise_for_disabled(context.disabled_tags)
        if hasattr(self, "render_to_output_async"):
            # pylint: disable=no-member
            return await self.render_to_output_async(context, buffer)
        return self.render_to_output(context, buffer)

    @abstractmethod
    def render_to_output(
        self,
        context: Context,
        buffer: TextIO,
    ) -> Optional[bool]:
        """Render this node to the output buffer."""

    async def render_to_output_async(
        self,
        context: Context,
        buffer: TextIO,
    ) -> Optional[bool]:
        """An async version of :meth:`liquid.ast.Node.render_to_output`."""
        return self.render_to_output(context, buffer)

    def children(self) -> List["ChildNode"]:
        """Return a list of child nodes and/or expressions associated with this node."""
        raise NotImplementedError(f"{self.__class__.__name__}.children")


class ChildNode(NamedTuple):
    """An AST node and expression pair with optional scope and load data.

    :param linenum: The line number of the child's first token in the template source
        text.
    :type linenum: int
    :param expression: An :class:`liquid.expression.Expression`. If not ``None``, this
        expression is expected to be related to the given :class:`liquid.ast.Node`.
    :type expression: Optional[liquid.expression.Expression]
    :param node: A :class:`liquid.ast.Node`. Typically a ``BlockNode`` or
        ``ConditionalBlockNode``.
    :type node: Optional[liquid.ast.Node]
    :param template_scope: A list of names the parent node adds to the template "local"
        scope. For example, the built-in ``assign``, ``capture``, ``increment`` and
        ``decrement`` tags all add names to the template scope. This helps us identify,
        through static analysis, names that are assumed to be "global".
    :type template_scope: Optional[List[str]]
    :param block_scope: A list of names available to the given child node. For example,
        the ``for`` tag adds the name "forloop" for the duration of its block.
    :type block_scope: Optional[List[str]]
    :param load_mode: If not ``None``, indicates that the given expression should be
        used to load a partial template. In "render" mode, the partial will be analyzed
        in an isolated namespace, without access to the parent's template local scope.
        In "include" mode, the partial will have access to the parents template local
        scope and the parent's scope can be updated by the partial template too.
    :type load_mode: Optional[Literal["render", "include"]]
    :param load_context: Meta data a template ``Loader`` might need to find the source
        of a partial template.
    """

    linenum: int
    expression: Optional[Expression] = None
    node: Optional[Node] = None
    template_scope: Optional[List[str]] = None
    block_scope: Optional[List[str]] = None
    load_mode: Optional[Literal["render", "include"]] = None
    load_context: Optional[Dict[str, str]] = None


class ParseTree(Node):
    """The root node of all syntax trees."""

    __slots__ = ("statements", "version")

    def __init__(self) -> None:
        self.statements: List[Node] = []

    def __str__(self) -> str:  # pragma: no cover
        return "".join(str(s) for s in self.statements)

    def __repr__(self) -> str:
        return f"ParseTree({self.statements})"

    # pylint: disable=useless-return
    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        for stmt in self.statements:
            stmt.render(context, buffer)
        return None


class IllegalNode(Node):
    """Parse tree node representing an illegal or unregistered tag.

    Illegal nodes are necessary when running in "warn" or "lax" mode. In
    strict mode, an exception should be raised as soon as an illegal token
    is found.
    """

    __slots__ = ("tok",)

    def __init__(self, tok: Token):
        self.tok = tok

    def __str__(self) -> str:
        return ""

    def __repr__(self) -> str:  # pragma: no cover
        return f"IllegalNode(tok={self.tok})"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        pass


class BlockNode(Node):
    """A parse tree node representing a sequence of statements."""

    __slots__ = ("tok", "statements", "forced_output")

    def __init__(self, tok: Token, statements: Optional[List[Node]] = None):
        self.tok = tok
        self.statements = statements or []
        self.forced_output = False

    def __str__(self) -> str:
        return "".join(str(s) for s in self.statements)

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        for stmt in self.statements:
            try:
                stmt.render(context, buffer)
            except Error as err:
                # Maybe resume rendering the block after an error.
                context.error(err)
        return True

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        for stmt in self.statements:
            try:
                await stmt.render_async(context, buffer)
            except Error as err:
                # Maybe resume rendering the block after an error.
                context.error(err)
        return True

    def children(self) -> List["ChildNode"]:
        return [
            ChildNode(
                linenum=self.tok.linenum,
                node=stmt,
            )
            for stmt in self.statements
        ]


class ConditionalBlockNode(Node):
    """A parse tree node representing a sequence of statements and a conditional
    expression."""

    __slots__ = ("tok", "condition", "block", "forced_output")

    def __init__(self, tok: Token, condition: Expression, block: BlockNode):
        self.tok = tok
        self.condition = condition
        self.block = block
        self.forced_output = block.forced_output

    def __str__(self) -> str:
        return f"{self.condition} {{ {self.block} }}"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        if self.condition.evaluate(context):
            self.block.render(context, buffer)
            return True
        return False

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        if await self.condition.evaluate_async(context):
            await self.block.render_async(context, buffer)
            return True
        return False

    def children(self) -> List["ChildNode"]:
        return [
            ChildNode(
                linenum=self.tok.linenum,
                node=self.block,
                expression=self.condition,
            )
        ]



"""Golden tests cases for testing liquid's built-in `default` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="nil",
        template=r"{{ nil | default: 'foo' }}",
        expect="foo",
    ),
    Case(
        description="false",
        template=r"{{ False | default: 'foo' }}",
        expect="foo",
    ),
    Case(
        description="empty string",
        template=r'{{ "" | default: "foo" }}',
        expect="foo",
    ),
    Case(
        description="empty array",
        template=r"{{ a | default: 'foo' }}",
        expect="foo",
        globals={"a": []},
    ),
    Case(
        description="empty object",
        template=r"{{ a | default: 'foo' }}",
        expect="foo",
        globals={"a": {}},
    ),
    Case(
        description="not empty string",
        template=r'{{ "hello" | default: "foo" }}',
        expect="hello",
    ),
    Case(
        description="not empty list",
        template=r'{{ a | default: "foo" | join: "#" }}',
        expect="hello#world",
        globals={"a": ["hello", "world"]},
    ),
    Case(
        description="not empty object",
        template=(
            r"{% assign b = a | default: foo %}"
            r"{% for item in b %}"
            r"({{ item[0] }},{{ item[1] }})"
            r"{% endfor %}"
        ),
        expect="(greeting,hello)",
        globals={"a": {"greeting": "hello"}, "foo": {"greeting": "goodbye"}},
    ),
    Case(
        description="too many arguments",
        template=r"{{ None | default: 'foo', 'bar', 'baz' }}",
        expect="",
        error=True,
    ),
    Case(
        description="missing argument",
        template=r"{{ false | default }}",
        expect="",
    ),
    Case(
        description="empty",
        template=r"{{ empty | default: bar }}",
        expect="",
    ),
    Case(
        description="allow false",
        template=r"{{ false | default: 'bar', allow_false:true }}",
        expect="false",
    ),
    Case(
        description="allow false from context",
        template=r"{{ false | default: 'bar', allow_false:foo }}",
        expect="false",
        globals={"foo": True},
    ),
    Case(
        description="undefined left value",
        template=r'{{ nosuchthing | default: "bar" }}',
        expect="bar",
    ),
    Case(
        description="true keyword argument before positional",
        template=r'{{ false | default: allow_false: true, "bar" }}',
        expect="false",
    ),
    Case(
        description="false keyword argument before positional",
        template=r'{{ false | default: allow_false: false, "bar" }}',
        expect="bar",
    ),
    Case(
        description="zero is not falsy",
        template=r'{{ 0 | default: "bar" }}',
        expect="0",
    ),
    Case(
        description="zero is not falsy with allow_false",
        template=r'{{ 0 | default: "bar", allow_false: true }}',
        expect="0",
    ),
    Case(
        description="0.0 is not falsy",
        template=r'{{ 0.0 | default: "bar" }}',
        expect="0.0",
    ),
]



"""Golden tests cases for testing liquid's built-in `size` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="size of an array",
        template=r"{{ a | size }}",
        globals={"a": ["a", "b", "c"]},
        expect="3",
    ),
    Case(
        description="size of a string",
        template=r"{{ a | size }}",
        globals={"a": "abc"},
        expect="3",
    ),
    Case(
        description="size of an empty array",
        template=r"{{ a | size }}",
        globals={"a": []},
        expect="0",
    ),
    Case(
        description="size of a hash",
        template=r"{{ a | size }}",
        globals={"a": {"a": 1, "b": 2}},
        expect="2",
    ),
    Case(
        description="unexpected argument",
        template=r"{{ a | size: 'foo' }}",
        globals={"a": [1, 2, 3]},
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | size }}",
        expect="0",
    ),
]



"""Golden tests cases for testing liquid's built-in `replace_last` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="replace substrings",
        template=(
            r'{{ "Take my protein pills and put my helmet on" '
            r'| replace_last: "my", "your" }}'
        ),
        expect="Take my protein pills and put your helmet on",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | replace_last: 'rain', 'foo' }}",
        expect="5",
    ),
    Case(
        description="argument not a string",
        template=r'{{ "hello5" | replace_last: 5, "your" }}',
        expect="helloyour",
    ),
    Case(
        description="missing argument",
        template=r'{{ "hello" | replace_last: "ll" }}',
        expect="",
        error=True,
    ),
    Case(
        description="missing arguments",
        template=r'{{ "hello" | replace_last }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello" | replace_last: "how", "are", "you" }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r'{{ nosuchthing | replace_last: "my", "your" }}',
        expect="",
    ),
    Case(
        description="undefined first argument",
        template=r'{{ "Take my protein" | replace_last: nosuchthing, "#" }}',
        expect="Take my protein#",
    ),
    Case(
        description="undefined second argument",
        template=(
            r'{{ "Take my protein pills and put my helmet on" '
            r'| replace_last: "my", nosuchthing }}'
        ),
        expect="Take my protein pills and put  helmet on",
    ),
]



"""Test cases for string filters."""
# pylint: disable=too-many-public-methods,too-many-lines,missing-class-docstring
import unittest

from functools import partial
from inspect import isclass

from typing import NamedTuple
from typing import Any
from typing import List
from typing import Dict

from liquid.environment import Environment
from liquid.exceptions import FilterArgumentError
from liquid.exceptions import FilterValueError
from liquid.exceptions import FilterError

from liquid.builtin.filters.string import capitalize
from liquid.builtin.filters.string import append
from liquid.builtin.filters.string import downcase
from liquid.builtin.filters.string import escape
from liquid.builtin.filters.string import escape_once
from liquid.builtin.filters.string import lstrip
from liquid.builtin.filters.string import newline_to_br
from liquid.builtin.filters.string import prepend
from liquid.builtin.filters.string import remove
from liquid.builtin.filters.string import remove_first
from liquid.builtin.filters.string import replace
from liquid.builtin.filters.string import replace_first
from liquid.builtin.filters.string import replace_last
from liquid.builtin.filters.string import slice_
from liquid.builtin.filters.string import split
from liquid.builtin.filters.string import upcase
from liquid.builtin.filters.string import strip
from liquid.builtin.filters.string import rstrip
from liquid.builtin.filters.string import strip_html
from liquid.builtin.filters.string import strip_newlines
from liquid.builtin.filters.string import truncate
from liquid.builtin.filters.string import truncatewords
from liquid.builtin.filters.string import url_encode
from liquid.builtin.filters.string import url_decode
from liquid.builtin.filters.string import base64_encode
from liquid.builtin.filters.string import base64_decode
from liquid.builtin.filters.string import base64_url_safe_encode
from liquid.builtin.filters.string import base64_url_safe_decode


class Case(NamedTuple):
    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class StringFilterTestCase(unittest.TestCase):
    """Test string filter functions."""

    def setUp(self) -> None:
        self.env = Environment()

    def _test(self, func, test_cases):
        if getattr(func, "with_environment", False):
            func = partial(func, environment=self.env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(
                    case.expect, (FilterArgumentError, FilterValueError, FilterError)
                ):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_capitalize(self):
        """Test capitalize filter function."""

        test_cases = [
            Case(
                description="lower case string",
                val="hello",
                args=[],
                kwargs={},
                expect="Hello",
            ),
            Case(
                description="already capitalized string",
                val="Hello",
                args=[],
                kwargs={},
                expect="Hello",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[2],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(capitalize, test_cases)

    def test_append(self):
        """Test append filter function."""

        test_cases = [
            Case(
                description="concat",
                val="hello",
                args=["there"],
                kwargs={},
                expect="hellothere",
            ),
            Case(
                description="not a string",
                val=5,
                args=["there"],
                kwargs={},
                expect="5there",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="hello5",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["hi"],
                kwargs={},
                expect="hi",
            ),
            Case(
                description="undefined argument",
                val="hi",
                args=[self.env.undefined("test")],
                kwargs={},
                expect="hi",
            ),
        ]

        self._test(append, test_cases)

    def test_downcase(self):
        """Test downcase filter function."""

        test_cases = [
            Case(
                description="make lower case",
                val="HELLO",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="HELLO",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(downcase, test_cases)

    def test_escape(self):
        """Test escape filter function."""

        test_cases = [
            Case(
                description="make HTML-safe",
                val="<p>test</p>",
                args=[],
                kwargs={},
                expect="&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="HELLO",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(escape, test_cases)

    def test_escape_once(self):
        """Test escape_once filter function."""

        test_cases = [
            Case(
                description="make HTML-safe",
                val="&lt;p&gt;test&lt;/p&gt;",
                args=[],
                kwargs={},
                expect="&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="make HTML-safe from mixed safe and markup.",
                val="&lt;p&gt;test&lt;/p&gt;<p>test</p>",
                args=[],
                kwargs={},
                expect="&lt;p&gt;test&lt;/p&gt;&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="HELLO",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(escape_once, test_cases)

    def test_lstrip(self):
        """Test lstrip filter function."""

        test_cases = [
            Case(
                description="left padded",
                val=" \t\r\n  hello",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="right padded",
                val="hello \t\r\n  ",
                args=[],
                kwargs={},
                expect="hello \t\r\n  ",
            ),
            Case(
                description="left and right padded",
                val=" \t\r\n  hello  \t\r\n ",
                args=[],
                kwargs={},
                expect="hello  \t\r\n ",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(lstrip, test_cases)

    def test_newline_to_br(self):
        """Test newline_to_br filter function."""

        test_cases = [
            Case(
                description="string with newlines",
                val="- apples\n- oranges\n",
                args=[],
                kwargs={},
                expect="- apples<br />\n- oranges<br />\n",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="reference implementation test 1",
                val="a\nb\nc",
                args=[],
                kwargs={},
                expect="a<br />\nb<br />\nc",
            ),
            Case(
                description="reference implementation test 2",
                val="a\r\nb\nc",
                args=[],
                kwargs={},
                expect="a<br />\nb<br />\nc",
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(newline_to_br, test_cases)

    def test_prepend(self):
        """Test prepend filter function."""

        test_cases = [
            Case(
                description="concat",
                val="hello",
                args=["there"],
                kwargs={},
                expect="therehello",
            ),
            Case(
                description="not a string",
                val=5,
                args=["there"],
                kwargs={},
                expect="there5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="5hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["hi"],
                kwargs={},
                expect="hi",
            ),
            Case(
                description="undefined argument",
                val="hi",
                args=[self.env.undefined("test")],
                kwargs={},
                expect="hi",
            ),
        ]

        self._test(prepend, test_cases)

    def test_remove(self):
        """Test remove filter function."""

        test_cases = [
            Case(
                description="remove substrings",
                val="I strained to see the train through the rain",
                args=["rain"],
                kwargs={},
                expect="I sted to see the t through the ",
            ),
            Case(
                description="not a string",
                val=5,
                args=["there"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["rain"],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined argument",
                val="I strained to see the train through the rain",
                args=[self.env.undefined("test")],
                kwargs={},
                expect="I strained to see the train through the rain",
            ),
        ]

        self._test(remove, test_cases)

    def test_remove_first(self):
        """Test remove_first filter function."""

        test_cases = [
            Case(
                description="remove substrings",
                val="I strained to see the train through the rain",
                args=["rain"],
                kwargs={},
                expect="I sted to see the train through the rain",
            ),
            Case(
                description="not a string",
                val=5,
                args=["rain"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["rain"],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined argument",
                val="I strained to see the train through the rain",
                args=[self.env.undefined("test")],
                kwargs={},
                expect="I strained to see the train through the rain",
            ),
        ]

        self._test(remove_first, test_cases)

    def test_replace(self):
        """Test replace filter function."""

        test_cases = [
            Case(
                description="replace substrings",
                val="Take my protein pills and put my helmet on",
                args=["my", "your"],
                kwargs={},
                expect="Take your protein pills and put your helmet on",
            ),
            Case(
                description="not a string",
                val=5,
                args=["rain", "foo"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5, "your"],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=["ll"],
                kwargs={},
                expect="heo",
            ),
            Case(
                description="missing arguments",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["my", "your"],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined first argument",
                val="Take my protein",
                args=[self.env.undefined("test"), "#"],
                kwargs={},
                expect="#T#a#k#e# #m#y# #p#r#o#t#e#i#n#",
            ),
            Case(
                description="undefined second argument",
                val="Take my protein pills and put my helmet on",
                args=["my", self.env.undefined("test")],
                kwargs={},
                expect="Take  protein pills and put  helmet on",
            ),
        ]

        self._test(replace, test_cases)

    def test_replace_first(self):
        """Test replace_first filter function."""
        test_cases = [
            Case(
                description="replace substrings",
                val="Take my protein pills and put my helmet on",
                args=["my", "your"],
                kwargs={},
                expect="Take your protein pills and put my helmet on",
            ),
            Case(
                description="not a string",
                val=5,
                args=["rain", "foo"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello5",
                args=[5, "your"],
                kwargs={},
                expect="helloyour",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=["ll"],
                kwargs={},
                expect="heo",
            ),
            Case(
                description="missing arguments",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["my", "your"],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined first argument",
                val="Take my protein pills and put my helmet on",
                args=[self.env.undefined("test"), "your"],
                kwargs={},
                expect="yourTake my protein pills and put my helmet on",
            ),
            Case(
                description="undefined second argument",
                val="Take my protein pills and put my helmet on",
                args=["my", self.env.undefined("test")],
                kwargs={},
                expect="Take  protein pills and put my helmet on",
            ),
        ]

        self._test(replace_first, test_cases)

    def test_replace_last(self):
        """Test replace_last filter function."""
        test_cases = [
            Case(
                description="not a string",
                val=5,
                args=["rain", "foo"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="replace substrings",
                val="Take my protein pills and put my helmet on",
                args=["my", "your"],
                kwargs={},
                expect="Take my protein pills and put your helmet on",
            ),
            Case(
                description="argument not a string",
                val="hello5",
                args=[5, "your"],
                kwargs={},
                expect="helloyour",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=["ll"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing arguments",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=["my", "your"],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined first argument",
                val="Take my protein pills and put my helmet on",
                args=[self.env.undefined("test"), "your"],
                kwargs={},
                expect="Take my protein pills and put my helmet onyour",
            ),
            Case(
                description="undefined second argument",
                val="Take my protein pills and put my helmet on",
                args=["my", self.env.undefined("test")],
                kwargs={},
                expect="Take my protein pills and put  helmet on",
            ),
        ]

        self._test(replace_last, test_cases)

    def test_slice(self):
        """Test slice filter function."""

        test_cases = [
            Case(
                description="zero",
                val="hello",
                args=[0],
                kwargs={},
                expect="h",
            ),
            Case(
                description="one",
                val="hello",
                args=[1],
                kwargs={},
                expect="e",
            ),
            Case(
                description="one length three",
                val="hello",
                args=[1, 3],
                kwargs={},
                expect="ell",
            ),
            Case(
                description="out of range",
                val="hello",
                args=[99],
                kwargs={},
                expect="",
            ),
            Case(
                description="not a string",
                val=5,
                args=[0],
                kwargs={},
                expect="5",
            ),
            Case(
                description="first argument not an integer",
                val="hello",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="second argument not an integer",
                val="hello",
                args=[5, "foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing arguments",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=[1, 2, 3],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="unexpected keyword arguments",
                val="hello",
                args=[1, 2],
                kwargs={"x": "y"},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[1, 3],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined first argument",
                val="hello",
                args=[self.env.undefined("test"), 3],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined second argument",
                val="hello",
                args=[1, self.env.undefined("test")],
                kwargs={},
                expect="e",
            ),
            Case(
                description="big negative second argument",
                val="foobar",
                args=[0, -(1 << 64)],
                kwargs={},
                expect="",
            ),
            Case(
                description="big positive second argument",
                val="foobar",
                args=[0, 1 << 63],
                kwargs={},
                expect="foobar",
            ),
            Case(
                description="big positive first argument",
                val="foobar",
                args=[1 << 63, 6],
                kwargs={},
                expect="",
            ),
            Case(
                description="big negative first argument",
                val="foobar",
                args=[-(1 << 63), 6],
                kwargs={},
                expect="",
            ),
            Case(
                description="array input",
                val=["f", "o", "o", "b", "a", "r"],
                args=[1, 3],
                kwargs={},
                expect=["o", "o", "b"],
            ),
        ]

        self._test(slice_, test_cases)

    def test_split(self):
        """Test split filter function."""

        test_cases = [
            Case(
                description="split string",
                val="Hi, how are you today?",
                args=[
                    " ",
                ],
                kwargs={},
                expect=["Hi,", "how", "are", "you", "today?"],
            ),
            Case(
                description="not a string",
                val=5,
                args=[" "],
                kwargs={},
                expect=["5"],
            ),
            Case(
                description="argument not a string",
                val="hello th1ere",
                args=[1],
                kwargs={},
                expect=["hello th", "ere"],
            ),
            Case(
                description="missing argument",
                val="hello there",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello there",
                args=[" ", ","],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[" "],
                kwargs={},
                expect=[""],
            ),
            Case(
                description="undefined argument",
                val="Hi, how are you today?",
                args=[self.env.undefined("test")],
                kwargs={},
                expect=[
                    "H",
                    "i",
                    ",",
                    " ",
                    "h",
                    "o",
                    "w",
                    " ",
                    "a",
                    "r",
                    "e",
                    " ",
                    "y",
                    "o",
                    "u",
                    " ",
                    "t",
                    "o",
                    "d",
                    "a",
                    "y",
                    "?",
                ],
            ),
        ]

        self._test(split, test_cases)

    def test_upcase(self):
        """Test upcase filter function."""

        test_cases = [
            Case(
                description="make lower case",
                val="hello",
                args=[],
                kwargs={},
                expect="HELLO",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(upcase, test_cases)

    def test_strip(self):
        """Test strip filter function."""

        test_cases = [
            Case(
                description="left padded",
                val=" \t\r\n  hello",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="right padded",
                val="hello \t\r\n  ",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="left and right padded",
                val=" \t\r\n  hello  \t\r\n ",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(strip, test_cases)

    def test_rstrip(self):
        """Test rstrip filter function."""

        test_cases = [
            Case(
                description="left padded",
                val=" \t\r\n  hello",
                args=[],
                kwargs={},
                expect=" \t\r\n  hello",
            ),
            Case(
                description="right padded",
                val="hello \t\r\n  ",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="left and right padded",
                val=" \t\r\n  hello  \t\r\n ",
                args=[],
                kwargs={},
                expect=" \t\r\n  hello",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(rstrip, test_cases)

    def test_strip_html(self):
        """Test strip_html filter function."""

        test_cases = [
            Case(
                description="some HTML markup",
                val="Have <em>you</em> read <strong>Ulysses</strong> &amp; &#20;?",
                args=[],
                kwargs={},
                expect="Have you read Ulysses &amp; &#20;?",
            ),
            Case(
                description="some HTML markup with HTML comment",
                val=(
                    "<!-- Have --><em>you</em> read "
                    "<strong>Ulysses</strong> &amp; &#20;?"
                ),
                args=[],
                kwargs={},
                expect="you read Ulysses &amp; &#20;?",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(strip_html, test_cases)

    def test_strip_newlines(self):
        """Test strip_newlines filter function."""

        test_cases = [
            Case(
                description="newline and other whitespace",
                val="hello there\nyou",
                args=[],
                kwargs={},
                expect="hello thereyou",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="reference implementation test 1",
                val="a\nb\nc",
                args=[],
                kwargs={},
                expect="abc",
            ),
            Case(
                description="reference implementation test 2",
                val="a\r\nb\nc",
                args=[],
                kwargs={},
                expect="abc",
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(strip_newlines, test_cases)

    def test_truncate(self):
        """Test truncate filter function."""

        test_cases = [
            Case(
                description="default end",
                val="Ground control to Major Tom.",
                args=[20],
                kwargs={},
                expect="Ground control to...",
            ),
            Case(
                description="custom end",
                val="Ground control to Major Tom.",
                args=[25, ", and so on"],
                kwargs={},
                expect="Ground control, and so on",
            ),
            Case(
                description="no end",
                val="Ground control to Major Tom.",
                args=[20, ""],
                kwargs={},
                expect="Ground control to Ma",
            ),
            Case(
                description="string is shorter than length",
                val="Ground control",
                args=[20],
                kwargs={},
                expect="Ground control",
            ),
            Case(
                description="not a string",
                val=5,
                args=[10],
                kwargs={},
                expect="5",
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=[5, "foo", "bar"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[5],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined first argument",
                val="Ground control to Major Tom.",
                args=[self.env.undefined("test")],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined second argument",
                val="Ground control to Major Tom.",
                args=[20, self.env.undefined("test")],
                kwargs={},
                expect="Ground control to Ma",
            ),
            Case(
                description="default length is 50",
                val="Ground control to Major Tom. Ground control to Major Tom.",
                args=[],
                kwargs={},
                expect="Ground control to Major Tom. Ground control to ...",
            ),
            Case(
                description="big positive argument",
                val="foobar",
                args=[1 << 63],
                kwargs={},
                expect="foobar",
            ),
            Case(
                description="big negative argument",
                val="foobar",
                args=[-(1 << 63)],
                kwargs={},
                expect="...",
            ),
        ]

        self._test(truncate, test_cases)

    def test_truncatewords(self):
        """Test truncatewords filter function."""

        test_cases = [
            Case(
                description="default end",
                val="Ground control to Major Tom.",
                args=[3],
                kwargs={},
                expect="Ground control to...",
            ),
            Case(
                description="custom end",
                val="Ground control to Major Tom.",
                args=[3, "--"],
                kwargs={},
                expect="Ground control to--",
            ),
            Case(
                description="no end",
                val="Ground control to Major Tom.",
                args=[3, ""],
                kwargs={},
                expect="Ground control to",
            ),
            Case(
                description="fewer words than word count",
                val="Ground control",
                args=[3],
                kwargs={},
                expect="Ground control",
            ),
            Case(
                description="not a string",
                val=5,
                args=[10],
                kwargs={},
                expect="5",
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=[5, "foo", "bar"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="reference implementation test 1",
                val="测试测试测试测试",
                args=[5],
                kwargs={},
                expect="测试测试测试测试",
            ),
            Case(
                description="reference implementation test 2",
                val="one two three",
                args=[2, 1],
                kwargs={},
                expect="one two1",
            ),
            Case(
                description="reference implementation test 3",
                val="one  two\tthree\nfour",
                args=[3],
                kwargs={},
                expect="one two three...",
            ),
            Case(
                description="reference implementation test 4",
                val="one two three four",
                args=[2],
                kwargs={},
                expect="one two...",
            ),
            Case(
                description="reference implementation test 5",
                val="one two three four",
                args=[0],
                kwargs={},
                expect="one...",
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[5],
                kwargs={},
                expect="",
            ),
            Case(
                description="undefined first argument",
                val="one two three four",
                args=[self.env.undefined("test")],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined second argument",
                val="one two three four",
                args=[2, self.env.undefined("test")],
                kwargs={},
                expect="one two",
            ),
            Case(
                description="very long argument",
                val="",
                args=[100000000000000],
                kwargs={},
                expect="",
            ),
            Case(
                description="default number of words is 15",
                val="a b c d e f g h i j k l m n o p q",
                args=[],
                kwargs={},
                expect="a b c d e f g h i j k l m n o...",
            ),
            Case(
                description="big positive argument",
                val="one two three four",
                args=[1 << 31],
                kwargs={},
                expect="one two three four",
            ),
            Case(
                description="big negative argument",
                val="one two three four",
                args=[-(1 << 31)],
                kwargs={},
                expect="one...",
            ),
        ]

        self._test(truncatewords, test_cases)

    def test_url_encode_html(self):
        """Test url_encode filter function."""

        test_cases = [
            Case(
                description="some special URL characters",
                val="email address is bob@example.com!",
                args=[],
                kwargs={},
                expect="email+address+is+bob%40example.com%21",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(url_encode, test_cases)

    def test_url_decode_html(self):
        """Test url_decode filter function."""

        test_cases = [
            Case(
                description="some special URL characters",
                val="email+address+is+bob%40example.com%21",
                args=[],
                kwargs={},
                expect="email address is bob@example.com!",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(url_decode, test_cases)

    def test_base64_encode(self):
        """Test base64_encode filter function."""

        test_cases = [
            Case(
                description="from string",
                val="_#/.",
                args=[],
                kwargs={},
                expect="XyMvLg==",
            ),
            Case(
                description="from string with URL unsafe",
                val=(
                    r"abcdefghijklmnopqrstuvwxyz "
                    r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
                    r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|"
                ),
                args=[],
                kwargs={},
                expect=(
                    "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1RVV"
                    "ldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8/Ljo7W117fVx8"
                ),
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="NQ==",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(base64_encode, test_cases)

    def test_base64_decode(self):
        """Test base64_decode filter function."""

        test_cases = [
            Case(
                description="from string",
                val="XyMvLg==",
                args=[],
                kwargs={},
                expect="_#/.",
            ),
            Case(
                description="from string with URL unsafe",
                val=(
                    "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1RVV"
                    "ldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8/Ljo7W117fVx8"
                ),
                args=[],
                kwargs={},
                expect=(
                    r"abcdefghijklmnopqrstuvwxyz "
                    r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
                    r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|"
                ),
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect=FilterError,
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(base64_decode, test_cases)

    def test_base64_url_safe_encode(self):
        """Test base64_url_safe_encode filter function."""

        test_cases = [
            Case(
                description="from string",
                val="_#/.",
                args=[],
                kwargs={},
                expect="XyMvLg==",
            ),
            Case(
                description="from string with URL unsafe",
                val=(
                    r"abcdefghijklmnopqrstuvwxyz "
                    r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
                    r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|"
                ),
                args=[],
                kwargs={},
                expect=(
                    "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1RVV"
                    "ldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8_Ljo7W117fVx8"
                ),
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="NQ==",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(base64_url_safe_encode, test_cases)

    def test_base64_url_safe_decode(self):
        """Test base64_url_safe_decode filter function."""

        test_cases = [
            Case(
                description="from string",
                val="XyMvLg==",
                args=[],
                kwargs={},
                expect="_#/.",
            ),
            Case(
                description="from string with URL unsafe",
                val=(
                    "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1RVV"
                    "ldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8_Ljo7W117fVx8"
                ),
                args=[],
                kwargs={},
                expect=(
                    r"abcdefghijklmnopqrstuvwxyz "
                    r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
                    r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|"
                ),
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect=FilterError,
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="undefined left value",
                val=self.env.undefined("test"),
                args=[],
                kwargs={},
                expect="",
            ),
        ]

        self._test(base64_url_safe_decode, test_cases)



"""Test liquid expressions."""
# pylint: disable=missing-class-docstring
from unittest import TestCase

from liquid import Context
from liquid import Environment

from liquid.expression import NIL
from liquid.expression import BLANK
from liquid.expression import CONTINUE
from liquid.expression import StringLiteral
from liquid.expression import RangeLiteral
from liquid.expression import IntegerLiteral
from liquid.expression import Identifier
from liquid.expression import IdentifierPathElement
from liquid.expression import PrefixExpression

from liquid.exceptions import LiquidTypeError


class CompareExpressionTestCase(TestCase):
    def test_compare_nil_to_none(self):
        """Test that nil is equal to None."""
        self.assertTrue(NIL.__eq__(None))  # pylint: disable=unnecessary-dunder-call

    def test_compare_blank_to_blank(self):
        """Test that blank is equal to blank."""
        self.assertTrue(BLANK.__eq__(BLANK))  # pylint: disable=unnecessary-dunder-call

    def test_compare_continue(self):
        """Test that continue is only equal to itself."""
        # pylint: disable=unnecessary-dunder-call
        self.assertFalse(CONTINUE.__eq__(object()))


class EvaluateExpressionTestCase(TestCase):
    def test_evaluate_continue(self):
        """Test that continue always evaluates to zero."""
        env = Environment()
        ctx = Context(env=env)

        self.assertEqual(CONTINUE.evaluate(ctx), 0)

    def test_evaluate_range_literal(self):
        """Test that ranges can not descend."""
        env = Environment()
        ctx = Context(env=env)

        rng = RangeLiteral(IntegerLiteral(9), IntegerLiteral(1))
        self.assertEqual(list(rng.evaluate(ctx)), [])

    def test_evaluate_prefix_expression(self):
        """Test that prefix expression raise a type error if the operator is
        unknown."""
        env = Environment()
        ctx = Context(env=env)

        expr = PrefixExpression("@", IntegerLiteral(1))
        with self.assertRaises(LiquidTypeError):
            expr.evaluate(ctx)


class HashExpressionTestCase(TestCase):
    def test_hash_literal(self):
        """Test that literals are hashable."""
        hash(StringLiteral("foo"))

    def test_hash_range(self):
        """Test that range literals are hashable."""
        rng = RangeLiteral(IntegerLiteral(9), IntegerLiteral(1))
        hash(rng)

    def test_hash_identifier(self):
        """Test that range identifiers are hashable."""
        ident = Identifier([IdentifierPathElement(0)])
        hash(ident)



"""Golden tests cases for testing liquid's built-in `slice` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="zero",
        template=r'{{ "hello" | slice: 0 }}',
        expect="h",
    ),
    Case(
        description="one",
        template=r'{{ "hello" | slice: 1 }}',
        expect="e",
    ),
    Case(
        description="one length three",
        template=r'{{ "hello" | slice: 1, 3 }}',
        expect="ell",
    ),
    Case(
        description="out of range",
        template=r'{{ "hello" | slice: 99 }}',
        expect="",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | slice: 1 }}",
        expect="",
    ),
    Case(
        description="first argument is a string",
        template=r'{{ "hello" | slice: "2" }}',
        expect="l",
    ),
    Case(
        description="first argument not an integer",
        template=r'{{ "hello" | slice: "foo" }}',
        expect="",
        error=True,
    ),
    Case(
        description="second argument not an integer",
        template=r'{{ "hello" | slice: 5, "foo" }}',
        expect="",
        error=True,
    ),
    Case(
        description="second argument is a string",
        template=r'{{ "hello" | slice: 3, "2" }}',
        expect="lo",
    ),
    Case(
        description="missing arguments",
        template=r'{{ "hello" | slice }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello" | slice: 1, 2, 3 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | slice: 1, 3 }}",
        expect="",
    ),
    Case(
        description="undefined first argument",
        template=r'{{ "hello" | slice: nosuchthing, 3 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined second argument",
        template=r'{{ "hello" | slice: 1, nosuchthing }}',
        expect="e",
    ),
    Case(
        description="slice an array of numbers",
        template=r"{{ a | slice: 2, 3 | join: '#' }}",
        expect="3#4#5",
        globals={"a": [1, 2, 3, 4, 5]},
    ),
    Case(
        description="first argument is a float",
        template=r"{{ 'Liquid' | slice: 2.2 }}",
        expect="",
        error=True,
    ),
    Case(
        description="second argument is a float",
        template=r"{{ 'Liquid' | slice: 1, 2.2 }}",
        expect="",
        error=True,
    ),
    Case(
        description="negative first argument",
        template=r"{{ 'Liquid' | slice: -2 }}",
        expect="i",
    ),
    Case(
        description="negative first argument and positive length",
        template=r"{{ 'Liquid' | slice: -2, 2 }}",
        expect="id",
    ),
    Case(
        description="negative first argument and negative length",
        template=r"{{ 'Liquid' | slice: -2, -1 }}",
        expect="",
    ),
    Case(
        description="negative first argument and length out of range",
        template=r"{{ 'Liquid' | slice: -2, 99 }}",
        expect="id",
    ),
]



"""Golden tests cases for testing liquid's built-in `base64_url_safe_encode` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="from string",
        template='{{ "_#/." | base64_url_safe_encode }}',
        expect="XyMvLg==",
    ),
    Case(
        description="from string with URL unsafe",
        template=r"{{ a | base64_url_safe_encode }}",
        globals={
            "a": (
                r"abcdefghijklmnopqrstuvwxyz "
                r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
                r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|"
            )
        },
        expect=(
            "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1RVV"
            "ldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8_Ljo7W117fVx8"
        ),
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | base64_url_safe_encode }}",
        expect="NQ==",
    ),
    Case(
        description="unexpected argument",
        template=r'{{ "hello" | base64_url_safe_encode: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | base64_url_safe_encode }}",
        expect="",
    ),
]



"""A JSON filter."""
import json

from typing import Any
from typing import Callable
from typing import Optional

from liquid.filter import int_arg
from liquid.filter import liquid_filter


class JSON:  # pylint: disable=too-few-public-methods
    """Serialize an object to a JSON formatted string.

    :param default: A function passed to `json.dumps`. This function is called
        in the event that the JSONEncoder does not know how to serialize an
        object. Defaults to ``None``.
    :type default: Optional[Callable[[Any], Any]]
    :param strip_tags: If ``True``, HTML tags will be removed from the resulting
        JSON formatted string. Defaults to ``True``.
    :type strip_tags: bool
    """

    name = "json"

    def __init__(self, default: Optional[Callable[[Any], Any]] = None):
        self.default = default

    @liquid_filter
    def __call__(
        self,
        obj: object,
        indent: Optional[object] = None,
    ) -> str:
        indent = int_arg(indent) if indent else None
        return json.dumps(obj, default=self.default, indent=indent)



"""Test cases for the `json` filter function."""
import unittest

from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import is_dataclass
from inspect import isclass

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple

from liquid.environment import Environment
from liquid.exceptions import Error
from liquid.exceptions import FilterArgumentError
from liquid.extra.filters._json import JSON


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


@dataclass
class MockData:
    """Mock data class."""

    length: int
    width: int


class JSONFilterTestCase(unittest.TestCase):
    """Test the JSON template filter."""

    def setUp(self) -> None:
        self.env = Environment()

    def _test(self, func, test_cases: Iterable[Case]):
        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(case.expect, Error):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_json_filter(self) -> None:
        """Test `json` filter function."""
        test_cases = [
            Case(
                description="serialize a string",
                val="hello",
                args=[],
                kwargs={},
                expect='"hello"',
            ),
            Case(
                description="serialize an int",
                val=42,
                args=[],
                kwargs={},
                expect="42",
            ),
            Case(
                description="serialize a dict with list",
                val={"foo": [1, 2, 3]},
                args=[],
                kwargs={},
                expect='{"foo": [1, 2, 3]}',
            ),
            Case(
                description="serialize an arbitrary object",
                val={"foo": MockData(3, 4)},
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="with indent",
                val={"foo": [1, 2, 3]},
                args=[4],
                kwargs={},
                expect='{\n    "foo": [\n        1,\n        2,\n        3\n    ]\n}',
            ),
        ]

        self._test(JSON(), test_cases)

    def test_json_with_encoder_func(self) -> None:
        """Test `json` filter function with default encoder."""
        test_cases = [
            Case(
                description="serialize a dataclass",
                val={"foo": MockData(3, 4)},
                args=[],
                kwargs={},
                expect=r'{"foo": {"length": 3, "width": 4}}',
            ),
            Case(
                description="serialize an arbitrary object",
                val={"foo": object()},
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        def default(obj: object) -> Dict[str, Any]:
            if is_dataclass(obj):
                return asdict(obj)
            raise TypeError(f"can't serialize object {obj}")

        self._test(JSON(default=default), test_cases)



"""Mick shop related filters."""
# pylint: disable=missing-function-docstring
from liquid.filter import liquid_filter
from liquid.filter import string_filter

from liquid.exceptions import FilterArgumentError


TRANSLATION_TABLE = {"'": None, '"': None, "(": None, ")": None, "[": None, "]": None}


def to_handle(val):
    val = val.lower().translate(TRANSLATION_TABLE)
    # val = re.sub(r"\W+", "-", val)
    val = "-".join(val.split())
    # if val.endswith("-"):
    #     val = val[:-1]
    # if val.startswith("-"):
    #     val = s[1:]

    return val


def _link_to(link, url, title=""):
    return f'<a href="{url}" title="{title}">{link}</a>'


@liquid_filter
def asset_url(val):
    return f"/files/1/[shop_id]/[shop_id]/assets/{val}"


@liquid_filter
def global_asset_url(val):
    return f"/global/{val}"


@liquid_filter
def shopify_asset_url(val):
    return f"/shopify/{val}"


@liquid_filter
def script_tag(val):
    return f'<script src="{val}" type="text/javascript"></script>'


@liquid_filter
def stylesheet_tag(val, media="all"):
    return f'<link href="{val}" rel="stylesheet" type="text/css"  media="{media}"  />'


@liquid_filter
def link_to(val, url, title=""):
    return _link_to(val, url, title)


@liquid_filter
def within(url, collection):
    return f"/collections/{collection['handle']}/{url}"


@liquid_filter
def img_tag(val, alt=""):
    return f'<img src="{val}" alt="{alt}" />'


@liquid_filter
def url_for_vendor(val):
    return f"/collections/{to_handle(val)}"


@liquid_filter
def url_for_type(val):
    return f"/collections/{to_handle(val)}"


@liquid_filter
def link_to_vendor(val):
    vendor = val
    if not vendor:
        return "Unknown Vendor"

    _url_for_vendor = f'"/collections/{to_handle(val)}"'
    return f'<a href="{_url_for_vendor}" title="{vendor}">{vendor}</a>'


@liquid_filter
def link_to_type(val):
    typ = val
    if not typ:
        return "Unknown Type"

    _url_for_type = f"/collections/{to_handle(val)}"
    return f'<a href="{_url_for_type}" title="{typ}">{typ}</a>'


@string_filter
def product_img_url(url, style="small"):
    if not url.startswith("products/"):
        raise FilterArgumentError(
            "product_img_url: can only be called on product images"
        )

    if style == "original":
        return f"/files/shops/random_number/{url}"

    if style in ("grande", "large", "medium", "compact", "small", "thumb", "icon"):
        # Does not attempt to mimic reference implementation URL.
        return f"/files/shops/random_number/products/{style}"

    raise FilterArgumentError(
        "product_img_url: valid parameters for filter 'size' are: "
        "original, grande, large, medium, compact, small, thumb and icon"
    )


@liquid_filter
def default_pagination(paginate):
    html = []

    if paginate["previous"]:
        link = _link_to(paginate["previous"]["title"], paginate["previous"]["url"])
        html.append(f'<span class="prev">{link}</span>')

    for part in paginate["parts"]:
        if part["is_link"]:
            link = _link_to(part["title"], part["url"])
            html.append(f'<span class="page">{link}</span>')
        elif part["title"] == paginate["current_page"]:  # Leading integer comparison.
            title = part["title"]
            html.append(f'<span class="page current">{title}</span>')
        else:
            title = part["title"]
            html.append(f'<span class="deco">{title}</span>')

    if paginate["next"]:
        link = _link_to(paginate["next"]["title"], paginate["next"]["url"])
        html.append(f'<span class="next">#{link}</span>')

    return " ".join(html)


@liquid_filter
def pluralize(val, singular, plural):
    if val == 1:
        return singular
    return plural



"""Golden tests cases for testing liquid's `assign` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="assign a filtered literal",
        template=r"{% assign foo = 'foo' | upcase %}{{ foo }}",
        expect="FOO",
    ),
    Case(
        description="local variables shadow global variables",
        template=r"{{ foo }}{% assign foo = 'foo' | upcase %}{{ foo }}",
        expect="barFOO",
        globals={"foo": "bar"},
    ),
    Case(
        description="assign a range literal",
        template=r"{% assign foo = (1..3) %}{{ foo | join: '#' }}",
        expect="1#2#3",
    ),
]



# pylint: disable=missing-module-docstring
from .lex import tokenize
from .parse import parse_call_arguments
from .parse import parse_keyword_arguments
from .parse import parse_macro_arguments

__all__ = (
    "parse_call_arguments",
    "parse_keyword_arguments",
    "parse_macro_arguments",
    "tokenize",
)



"""Golden tests cases for testing liquid's built-in `abs` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="positive integer",
        template=r"{{ 5 | abs }}",
        expect="5",
    ),
    Case(
        description="negative integer",
        template=r"{{ -5 | abs }}",
        expect="5",
    ),
    Case(
        description="positive float",
        template=r"{{ 5.4 | abs }}",
        expect="5.4",
    ),
    Case(
        description="negative float",
        template=r"{{ -5.4 | abs }}",
        expect="5.4",
    ),
    Case(
        description="zero",
        template=r"{{ 0 | abs }}",
        expect="0",
    ),
    Case(
        description="positive string integer",
        template=r"{{ '5' | abs }}",
        expect="5",
    ),
    Case(
        description="negative string integer",
        template=r"{{ '-5' | abs }}",
        expect="5",
    ),
    Case(
        description="positive string float",
        template=r"{{ '5.1' | abs }}",
        expect="5.1",
    ),
    Case(
        description="negative string float",
        template=r"{{ '-5.1' | abs }}",
        expect="5.1",
    ),
    Case(
        description="unexpected argument",
        template=r"{{ -3 | abs: 1 }}",
        expect="",
        error=True,
    ),
    Case(
        description="string not a number",
        template=r"{{ 'hello' | abs }}",
        expect="0",
    ),
    Case(
        description="not a string, int or float",
        template=r"{{ a | abs }}",
        expect="0",
        globals={"a": {}},
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | abs }}",
        expect="0",
    ),
]



"""Tag and node definition for the built-in "assign" tag."""
import re
import sys

from typing import List
from typing import Optional
from typing import TextIO

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION

from liquid.ast import ChildNode
from liquid.ast import Node
from liquid.expression import Expression
from liquid.tag import Tag
from liquid.context import Context
from liquid.stream import TokenStream
from liquid.expression import AssignmentExpression
from liquid.exceptions import LiquidSyntaxError

from liquid.parse import expect
from liquid.expressions.common import ASSIGN_IDENTIFIER_PATTERN

RE_ASSIGNMENT = re.compile(rf"^({ASSIGN_IDENTIFIER_PATTERN})\s*=\s*(.+)$")

TAG_ASSIGN = sys.intern("assign")


class AssignNode(Node):
    """Parse tree node for the built-in "assign" tag."""

    __slots__ = ("tok", "expression")

    def __init__(self, tok: Token, expression: AssignmentExpression):
        self.tok = tok
        self.expression = expression

    def __str__(self) -> str:
        return f"var ({self.expression})"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        self.expression.evaluate(context)
        return False

    async def render_to_output_async(
        self, context: Context, _: TextIO
    ) -> Optional[bool]:
        await self.expression.evaluate_async(context)
        return False

    def children(self) -> List[ChildNode]:
        return [
            ChildNode(
                linenum=self.tok.linenum,
                expression=self.expression,
                template_scope=[self.expression.name],
            )
        ]


class AssignTag(Tag):
    """The built-in cycle tag."""

    name = TAG_ASSIGN
    block = False

    def _parse_expression(self, value: str) -> Expression:
        return self.env.parse_filtered_expression_value(value)

    def parse(self, stream: TokenStream) -> AssignNode:
        expect(stream, TOKEN_TAG, value=TAG_ASSIGN)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)

        match = RE_ASSIGNMENT.match(stream.current.value)
        if match:
            name, expression = match.groups()
        else:
            raise LiquidSyntaxError(
                f'invalid assignment expression "{stream.current.value}"',
                linenum=stream.current.linenum,
            )

        return AssignNode(
            tok,
            AssignmentExpression(name, self._parse_expression(expression)),
        )



"""Golden tests cases for testing liquid's built-in `concat` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="range literal concat filter left value",
        template=r"{{ (1..3) | concat: foo | join: '#' }}",
        expect="1#2#3#5#6#7",
        globals={"foo": [5, 6, 7]},
    ),
    Case(
        description="two arrays of strings",
        template=r"{{ a | concat: b | join: '#' }}",
        expect="a#b#c#d",
        globals={"a": ["a", "b"], "b": ["c", "d"]},
    ),
    Case(
        description="missing argument is an error",
        template=r"{{ a | concat | join: '#' }}",
        expect="a#b",
        globals={"a": ["a", "b"]},
        error=True,
    ),
    Case(
        description="non array-like argument is an error",
        template=r"{{ a | concat: b | join: '#' }}",
        expect="",
        globals={"a": ["a", "b"], "b": 5},
        error=True,
    ),
    Case(
        description="left value is not array-like",
        template=r"{{ a | concat: b | join: '#' }}",
        expect="ab#c#d",
        globals={"a": "ab", "b": ["c", "d"]},
    ),
    Case(
        description="left value contains non string",
        template=r"{{ a | concat: b | join: '#' }}",
        expect="a#b#5#c#d",
        globals={"a": ["a", "b", 5], "b": ["c", "d"]},
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | concat: b | join: '#' }}",
        expect="c#d",
        globals={"b": ["c", "d"]},
    ),
    Case(
        description="undefined argument is an error",
        template=r"{{ a | concat: nosuchthing | join: '#' }}",
        expect="",
        globals={"a": ["a", "b"]},
        error=True,
    ),
    Case(
        description="nested left value gets flattened",
        template=r"{{ a | concat: b | join: '#' }}",
        expect="a#x#b#y#z#c#d",
        globals={"a": [["a", "x"], ["b", ["y", ["z"]]]], "b": ["c", "d"]},
    ),
]



"""Golden tests cases for testing liquid's `unless` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="literal false condition",
        template=r"{% unless false %}foo{% endunless %}",
        expect="foo",
    ),
    Case(
        description="literal true condition",
        template=r"{% unless true %}foo{% endunless %}",
        expect="",
    ),
    Case(
        description="blocks that contain only whitespace are not rendered",
        template=r"{% unless false %}  {% endunless %}",
        expect="",
    ),
    Case(
        description="alternative block",
        template=r"{% unless true %}foo{% else %}bar{% endunless %}",
        expect="bar",
    ),
    Case(
        description="conditional alternative block",
        template=r"{% unless true %}foo{% elsif true %}bar{% endunless %}",
        expect="bar",
    ),
    Case(
        description="conditional alternative block with default",
        template=(
            r"{% unless true %}foo"
            r"{% elsif false %}bar"
            r"{% else %}hello"
            r"{% endunless %}"
        ),
        expect="hello",
    ),
    Case(
        description=("zero is not equal to false"),
        template=(r"{% unless 0 == false %}Hello{% else %}Goodbye{% endunless %}"),
        expect="Hello",
    ),
    Case(
        description=("zero is truthy"),
        template=(r"{% unless 0 %}Hello{% else %}Goodbye{% endunless %}"),
        expect="Goodbye",
    ),
    Case(
        description=("one is not equal to true"),
        template=(r"{% unless 1 == true %}Hello{% else %}Goodbye{% endunless %}"),
        expect="Hello",
    ),
]



"""Golden test cases for testing Liquid's special, built-in properties."""

from liquid.golden.case import Case

cases = [
    Case(
        description="first of an array",
        template=r"{{ a.first }}",
        expect="3",
        globals={"a": [3, 2, 1]},
    ),
    Case(
        description="last of an array",
        template=r"{{ a.last }}",
        expect="1",
        globals={"a": [3, 2, 1]},
    ),
    Case(
        description="size of an array",
        template=r"{{ a.size }}",
        expect="3",
        globals={"a": [3, 2, 1]},
    ),
    Case(
        description="size of a string",
        template=r"{{ s.size }}",
        expect="5",
        globals={"s": "hello"},
    ),
    Case(
        description="first of a string",
        template=r"{{ s.first }}",
        expect="",
        globals={"s": "hello"},
    ),
    Case(
        description="first of an object",
        template=r"{{ obj.first | join: '#' }}",
        expect="a#1",
        globals={"obj": {"a": 1, "b": 2}},
    ),
    Case(
        description="first of an empty object",
        template=r"{{ obj.first | join: '#' }}",
        expect="",
        globals={"obj": {}},
    ),
    Case(
        description="last of a string",
        template=r"{{ s.last }}",
        expect="",
        globals={"s": "hello"},
    ),
    Case(
        description="last of a object",
        template=r"{{ obj.last }}",
        expect="",
        globals={"obj": {"a": 1, "b": 2}},
    ),
    Case(
        description="size of an object with a size property",
        template=r"{{ obj.size }}",
        expect="99",
        globals={"obj": {"size": 99}},
    ),
    Case(
        description="first of an object with a first property",
        template=r"{{ obj.first }}",
        expect="99",
        globals={"obj": {"a": 1, "first": 99}},
    ),
    Case(
        description="last of an object with a last property",
        template=r"{{ obj.last }}",
        expect="99",
        globals={"obj": {"a": 1, "last": 99, "b": 42}},
    ),
    Case(
        description="size of undefined",
        template=r"{{ nosuchthing.last }}",
        expect="",
        globals={},
    ),
]



"""Golden tests cases for testing liquid's `comment` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="don't render comments",
        template=r"{% comment %}foo{% endcomment %}",
        expect="",
    ),
    Case(
        description="respect whitespace control in comments",
        template="\n{%- comment %}foo{% endcomment -%}\t \r",
        expect="",
    ),
    Case(
        description="don't render comments with tags",
        template=(
            r"{% comment %}"
            r"{% if true %}"
            r"{{ title }}"
            r"{% endif %}"
            r"{% endcomment %}"
        ),
        expect="",
    ),
]
